from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class Opportunity(Resource):
    """
    https://dev.chartmogul.com/reference/opportunities
    """

    _path = "/opportunities{/uuid}"
    _root_key = "entries"
    _many = namedtuple("Opportunities", [_root_key, "has_more", "cursor"])

    class _Schema(Schema):

        uuid = fields.String()
        customer_uuid = fields.String()
        owner = fields.String()
        pipeline = fields.String()
        pipeline_stage = fields.String()
        estimated_close_date = fields.Date()
        currency = fields.String()
        amount_in_cents = fields.Int()
        type = fields.String()
        forecast_category = fields.String()
        win_likelihood = fields.Int()
        custom = fields.Dict(allow_none=True)
        created_at = fields.DateTime()
        updated_at = fields.DateTime()

        @post_load
        def make(self, data, **kwargs):
            return Opportunity(**data)

    _schema = _Schema(unknown=EXCLUDE)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class CustomerNote(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#customer-notes
    """

    _path = "/customer_notes{/uuid}"
    _root_key = "entries"
    _many = namedtuple("CustomerNotes", [_root_key, "has_more", "cursor"])

    class _Schema(Schema):
        uuid = fields.String()
        customer_uuid = fields.String()
        type = fields.String()
        text = fields.String()
        author = fields.String()
        call_duration = fields.Int()
        created_at = fields.DateTime()
        updated_at = fields.DateTime()

        @post_load
        def make(self, data, **kwargs):
            return CustomerNote(**data)

    _schema = _Schema(unknown=EXCLUDE)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class PlanGroupPlans(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#plan_groups
    """

    _path = "/plan_groups{/uuid}/plans"
    _root_key = "plans"
    _many = namedtuple("PlanGroupPlans", [_root_key, "has_more", "cursor"], defaults=[None, None])

    class _Schema(Schema):
        uuid = fields.String()
        data_source_uuid = fields.String()
        name = fields.String()
        interval_count = fields.Int()
        interval_unit = fields.String()
        external_id = fields.String()

        @post_load
        def make(self, data, **kwargs):
            return PlanGroupPlans(**data)

    _schema = _Schema(unknown=EXCLUDE)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class Task(Resource):
    """
    https://dev.chartmogul.com/reference/tasks
    """

    _path = "/tasks{/uuid}"
    _root_key = "entries"
    _many = namedtuple("Tasks", [_root_key, "has_more", "cursor"])

    class _Schema(Schema):

        task_uuid = fields.String()
        customer_uuid = fields.String()
        assignee = fields.String()
        task_details = fields.String()
        due_date = fields.DateTime()
        completed_at = fields.DateTime(allow_none=True)
        created_at = fields.DateTime()
        updated_at = fields.DateTime()

        @post_load
        def make(self, data, **kwargs):
            return Task(**data)

    _schema = _Schema(unknown=EXCLUDE)



from marshmallow import Schema, fields, post_load, EXCLUDE, INCLUDE
from ..resource import Resource, DataObject
from chartmogul import ArgumentMissingError
from collections import namedtuple


class ExportParams(DataObject):
    class _Schema(Schema):
        kind = fields.String()
        params = fields.Dict(allow_none=True)

        @post_load
        def make(self, data, **kwargs):
            return ExportParams(**data)


class ActivitiesExport(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#
    """

    _path = "/activities_export"
    _many = namedtuple("ActivitiesExport", ["id", "status", "file_url", "expires_at", "created_at"])

    class _Schema(Schema):
        # Create
        start_date = fields.DateTime()
        end_date = fields.DateTime()
        type = fields.String()

        # Retrieve
        id = fields.String()
        status = fields.String()
        file_url = fields.String(allow_none=True)
        params = fields.Nested(ExportParams._Schema(unknown=INCLUDE), allow_none=True)
        expires_at = fields.DateTime(allow_none=True)
        created_at = fields.DateTime()

        @post_load
        def make(self, data, **kwargs):
            return ActivitiesExport(**data)

    _schema = _Schema(unknown=EXCLUDE)

    @classmethod
    def _validate_arguments(cls, method, kwargs):
        # This enforces user to pass correct argument
        if method in ["retrieve"] and "id" not in kwargs:
            raise ArgumentMissingError(
                "Please pass 'id' parameter to retrieve your export request status"
            )
        if method in ["create"] and "data" not in kwargs:
            raise ArgumentMissingError("Please pass 'data' parameter")


ActivitiesExport.create = ActivitiesExport._method("create", "post", "/activities_export")
ActivitiesExport.retrieve = ActivitiesExport._method("retrieve", "get", "/activities_export/{id}")



import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

METHOD_WHITELIST = ["HEAD", "GET", "POST", "PUT", "DELETE", "OPTIONS", "TRACE"]
STATUS_FORCELIST = (429, 500, 502, 503, 504, 520, 524)


def requests_retry_session(
    retries=20,
    backoff_factor=2,
    session=None,
):
    session = session or requests.Session()
    adapter = _retry_adapter(retries, backoff_factor)
    session.mount("https://", adapter)
    return session


def _retry_adapter(retries, backoff_factor):
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        status=retries,
        allowed_methods=METHOD_WHITELIST,
        status_forcelist=STATUS_FORCELIST,
        backoff_factor=backoff_factor,
    )
    return HTTPAdapter(max_retries=retry)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class SubscriptionEvent(Resource):
    """
    https://dev.chartmogul.com/reference/subscription-events
    """

    _path = "/subscription_events"
    _root_key = "subscription_events"
    _many = namedtuple(
        "SubscriptionEvents", [_root_key, "has_more", "cursor"], defaults=[None, None]
    )

    class _Schema(Schema):
        id = fields.Int(required=True)
        data_source_uuid = fields.String(required=True)
        customer_external_id = fields.String(required=True)
        event_type = fields.String(required=True)
        event_date = fields.DateTime(required=True)
        effective_date = fields.DateTime(required=True)
        subscription_external_id = fields.String(allow_none=True)
        plan_external_id = fields.String(allow_none=True)
        currency = fields.String(allow_none=True)
        amount_in_cents = fields.Int(allow_none=True)
        quantity = fields.Int(allow_none=True)
        subscription_set_external_id = fields.String(allow_none=True)
        tax_amount_in_cents = fields.Int(allow_none=True)
        retracted_event_id = fields.String(allow_none=True)
        external_id = fields.String(allow_none=True)
        event_order = fields.Int(allow_none=True)

        @post_load
        def make(self, data, **kwargs):
            return SubscriptionEvent(**data)

    _schema = _Schema(unknown=EXCLUDE)


SubscriptionEvent.all = SubscriptionEvent._method("all", "get", "/subscription_events")
SubscriptionEvent.destroy_with_params = SubscriptionEvent._method(
    "destroy_with_params", "delete", "/subscription_events"
)
SubscriptionEvent.modify_with_params = SubscriptionEvent._method(
    "modify_with_params", "patch", "/subscription_events"
)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class DataSource(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#data-sources
    """

    _path = "/data_sources{/uuid}"
    _root_key = "data_sources"
    _many = namedtuple("DataSources", [_root_key])

    class _Schema(Schema):
        uuid = fields.String()
        name = fields.String()
        created_at = fields.DateTime()
        status = fields.Str()
        system = fields.Str()

        @post_load
        def make(self, data, **kwargs):
            return DataSource(**data)

    _schema = _Schema(unknown=EXCLUDE)



from marshmallow import Schema, fields, post_load, EXCLUDE
from chartmogul.resource import Resource
from collections import namedtuple


class CustomerSubscription(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#list-customer-subscriptions
    https://dev.chartmogul.com/v1.0/reference#list-a-customers-subscriptions
    """

    _path = "/customers{/uuid}/subscriptions"
    _root_key = "entries"
    _many = namedtuple("Subscriptions", [_root_key, "has_more", "cursor"], defaults=[None, None])

    class _Schema(Schema):
        id = fields.Int(allow_none=True)
        plan = fields.String(allow_none=True)
        quantity = fields.Int(allow_none=True)
        mrr = fields.Int(allow_none=True)
        arr = fields.Int(allow_none=True)
        status = fields.String(allow_none=True)
        billing_cycle = fields.String(data_key="billing-cycle", allow_none=True)
        billing_cycle_count = fields.Int(data_key="billing-cycle-count", allow_none=True)
        start_date = fields.DateTime(data_key="start-date", allow_none=True)
        end_date = fields.DateTime(data_key="end-date", allow_none=True)
        currency = fields.String(allow_none=True)
        currency_sign = fields.String(data_key="currency-sign", allow_none=True)

        # /import namespace
        uuid = fields.String(allow_none=True)
        external_id = fields.String(allow_none=True)
        subscription_set_external_id = fields.String(allow_none=True)
        plan_uuid = fields.String(allow_none=True)
        customer_uuid = fields.String(allow_none=True)
        data_source_uuid = fields.String(allow_none=True)
        cancellation_dates = fields.List(fields.DateTime(), allow_none=True)

        @post_load
        def make(self, data, **kwargs):
            return CustomerSubscription(**data)

    _schema = _Schema(unknown=EXCLUDE)

    # /import has different paging
    @classmethod
    def _loadJSON(cls, jsonObj):
        if "subscriptions" in jsonObj:
            _many = namedtuple(
                "Subscriptions",
                ["subscriptions", "has_more", "cursor", "customer_uuid"],
            )
            return _many(
                cls._schema.load(jsonObj["subscriptions"], many=True),
                customer_uuid=jsonObj.get("customer_uuid", None),
                has_more=jsonObj.get("has_more", None),
                cursor=jsonObj.get("cursor", None),
            )
        else:
            return super(CustomerSubscription, cls)._loadJSON(jsonObj)


# /import namespace
CustomerSubscription.list_imported = CustomerSubscription._method(
    "list_imported", "get", "/import/customers{/uuid}/subscriptions"
)
CustomerSubscription.cancel = CustomerSubscription._method(
    "cancel", "patch", "/import/subscriptions{/uuid}"
)
CustomerSubscription.modify = CustomerSubscription._method(
    "modify", "patch", "/import/subscriptions{/uuid}"
)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource, DataObject, _add_method
from collections import namedtuple

metrics = ["customers", "customer-churn-rate", "arr", "asp", "mrr", "arpa", "mrr-churn-rate", "ltv"]


class Summary(DataObject):
    """
    Optional information about a series of metrics.
    """

    class _Schema(Schema):
        current = fields.Float()
        previous = fields.Float()
        percentage_change = fields.Float(data_key="percentage-change")
        # All metrics percentage change
        for metric in metrics:
            pc = metric + "-percentage-change"
            current_pc = "current-" + metric
            previous_pc = "previous-" + metric
            locals()[pc.replace("-", "_")] = fields.Float(data_key=pc)
            locals()[current_pc.replace("-", "_")] = fields.Float(data_key=current_pc)
            locals()[previous_pc.replace("-", "_")] = fields.Float(data_key=previous_pc)

        @post_load
        def make(self, data, **kwargs):
            return Summary(**data)

    _schema = _Schema(unknown=EXCLUDE)


class Metrics(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#introduction-metrics-api
    """

    _path = "/metrics/all"
    _root_key = "entries"
    _many_cls = namedtuple("Metrics", [_root_key, "summary"])
    _many_cls.__new__.__defaults__ = (None,) * len(_many_cls._fields)

    class _Schema(Schema):
        """
        Fields are optional, so a subset present is good enough
        """

        date = fields.Date()
        customer_churn_rate = fields.Float(data_key="customer-churn-rate")
        mrr_churn_rate = fields.Float(data_key="mrr-churn-rate")
        ltv = fields.Float()
        customers = fields.Int()
        asp = fields.Int()
        arpa = fields.Int()
        arr = fields.Int()
        mrr = fields.Int()
        # MRR only
        mrr_new_business = fields.Int(data_key="mrr-new-business")
        mrr_expansion = fields.Int(data_key="mrr-expansion")
        mrr_contraction = fields.Int(data_key="mrr-contraction")
        mrr_churn = fields.Int(data_key="mrr-churn")
        mrr_reactivation = fields.Int(data_key="mrr-reactivation")
        percentage_change = fields.Float(data_key="percentage-change")
        for metric in metrics:
            pc = metric + "-percentage-change"
            locals()[pc.replace("-", "_")] = fields.Float(data_key=pc)

        @post_load
        def make(self, data, **kwargs):
            return Metrics(**data)

    _schema = _Schema(unknown=EXCLUDE)

    @classmethod
    def _many(cls, entries, **kwargs):
        if "summary" in kwargs:
            kwargs["summary"] = Summary._schema.load(kwargs["summary"])
        return cls._many_cls(entries, **kwargs)


_add_method(Metrics, "mrr", "get", path="/metrics/mrr")
_add_method(Metrics, "arr", "get", path="/metrics/arr")
_add_method(Metrics, "arpa", "get", path="/metrics/arpa")
_add_method(Metrics, "asp", "get", path="/metrics/asp")
_add_method(Metrics, "customer_count", "get", path="/metrics/customer-count")
_add_method(Metrics, "customer_churn_rate", "get", path="/metrics/customer-churn-rate")
_add_method(Metrics, "mrr_churn_rate", "get", path="/metrics/mrr-churn-rate")
_add_method(Metrics, "ltv", "get", path="/metrics/ltv")



from marshmallow import Schema, fields, post_load, EXCLUDE
from chartmogul.resource import Resource
from collections import namedtuple


class CustomerActivity(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#list-customer-subscriptions
    """

    _path = "/customers{/uuid}/activities"
    _root_key = "entries"
    _many = namedtuple("Activities", [_root_key, "has_more", "cursor"], defaults=[None, None])

    class _Schema(Schema):
        id = fields.Int()
        activity_arr = fields.Int(data_key="activity-arr")
        activity_mrr = fields.Int(data_key="activity-mrr")
        activity_mrr_movement = fields.Int(data_key="activity-mrr-movement")
        currency = fields.String()
        currency_sign = fields.String(data_key="currency-sign")
        date = fields.DateTime()
        description = fields.String()
        type = fields.String()
        subscription_external_id = fields.String(data_key="subscription-external-id")

        @post_load
        def make(self, data, **kwargs):
            return CustomerActivity(**data)

    _schema = _Schema(unknown=EXCLUDE)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class Contact(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#contacts
    """

    _path = "/contacts{/uuid}"
    _root_key = "entries"
    _many = namedtuple("Contacts", [_root_key, "has_more", "cursor"])

    class _Schema(Schema):
        uuid = fields.String()
        customer_uuid = fields.String(allow_none=True)
        data_source_uuid = fields.String(allow_none=True)
        customer_external_id = fields.String(allow_none=True)
        first_name = fields.String(allow_none=True)
        last_name = fields.String(allow_none=True)
        position = fields.Int(allow_none=True)
        email = fields.String(allow_none=True)
        title = fields.String(allow_none=True)
        notes = fields.String(allow_none=True)
        phone = fields.String(allow_none=True)
        linked_in = fields.String(allow_none=True)
        twitter = fields.String(allow_none=True)
        custom = fields.Dict(allow_none=True)

        @post_load
        def make(self, data, **kwargs):
            return Contact(**data)

    _schema = _Schema(unknown=EXCLUDE)


Contact.merge = Contact._method("merge", "post", "/contacts/{into_uuid}/merge/{from_uuid}")



#!/usr/bin/env python
import os
import re
import sys

from codecs import open
from setuptools import setup

github_url = "https://github.com/chartmogul/chartmogul-python"

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    sys.exit()

requires = [
    "requests>=2.31.0",
    "uritemplate>=4.1.1",
    "promise>=2.3.0",
    "marshmallow>=3.24.0,<5.0.0",
    "urllib3>=2.2.2",
]
test_requirements = [
    "mock>=5.1.0",
    "requests-mock>=1.11.0",
    "PyYAML>=6.0.1",
    "httpretty>=1.1.4",
    "wrapt>=1.15.0",
]

with open("chartmogul/version.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

setup(
    name="chartmogul",
    version=version,
    description="Python library for ChartMogul API.",
    long_description="`See documentation on GitHub <" + github_url + ">`_",
    author="Petr Kopac (ChartMogul Ltd.)",
    author_email="petr@chartmogul.com",
    url="https://chartmogul.com",
    download_url=github_url + "/tarball/v" + version,
    packages=[
        "chartmogul",
        "chartmogul.api",
        "chartmogul.api.customers",
    ],
    package_data={"": ["LICENSE", "NOTICE"], "chartmogul": ["*.pem"]},
    package_dir={
        "chartmogul": "chartmogul",
        "chartmogul.api": "chartmogul/api",
        "chartmogul.api.customers": "chartmogul/api/customers",
    },
    include_package_data=True,
    install_requires=requires,
    license="MIT",
    zip_safe=False,
    extras_require={
        "testing": test_requirements
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from collections import namedtuple


class Activity(Resource):
    """
    https://dev.chartmogul.com/reference/list-activities
    """

    _path = "/activities"
    _root_key = "entries"
    _many = namedtuple(
        "Activities", [_root_key, "has_more", "per_page", "cursor"], defaults=[None, None, None]
    )

    class _Schema(Schema):
        activity_arr = fields.Int(data_key="activity-arr")
        activity_mrr = fields.Int(data_key="activity-mrr")
        activity_mrr_movement = fields.Int(data_key="activity-mrr-movement")
        currency = fields.String()
        date = fields.DateTime()
        description = fields.String()
        type = fields.String()
        subscription_external_id = fields.String(data_key="subscription-external-id")
        subscription_set_external_id = fields.String(data_key="subscription-set-external-id")
        plan_external_id = fields.String(data_key="plan-external-id")
        customer_name = fields.String(data_key="customer-name")
        customer_uuid = fields.String(data_key="customer-uuid")
        customer_external_id = fields.String(data_key="customer-external-id")
        billing_connector_uuid = fields.String(data_key="billing-connector-uuid")
        uuid = fields.String(data_key="uuid")

        @post_load
        def make(self, data, **kwargs):
            return Activity(**data)

    _schema = _Schema(unknown=EXCLUDE)



# -*- coding: utf-8 -*-
from .api.config import Config
from .errors import (
    APIError,
    ConfigurationError,
    ArgumentMissingError,
    DeprecatedArgumentError,
)

from .api.customers.activity import CustomerActivity
from .api.attributes import Attributes
from .api.custom_attrs import CustomAttributes
from .api.customer import Customer
from .api.customer_note import CustomerNote
from .api.contact import Contact
from .api.data_source import DataSource
from .api.invoice import Invoice
from .api.metrics import Metrics
from .api.opportunity import Opportunity
from .api.task import Task
from .api.ping import Ping
from .api.plan import Plan
from .api.plan_group import PlanGroup
from .api.subscription_event import SubscriptionEvent
from .api.customers.subscription import CustomerSubscription
from .api.customers.subscription import CustomerSubscription as Subscription
from .api.tags import Tags
from .api.transaction import Transaction
from .api.account import Account
from .api.activity import Activity
from .api.activities_export import ActivitiesExport

from .version import __version__


"""
ChartMogul API Python Client

Provides convenient Python bindings for ChartMogul's API.

:copyright: (c) 2023 by ChartMogul Ltd.
:license: MIT, see LICENSE for more details.
"""

__title__ = "chartmogul"
__build__ = 0x000000
__author__ = "ChartMogul Ltd"
__license__ = "MIT"
__copyright__ = "Copyright 2023 ChartMogul Ltd"



from marshmallow import Schema, fields, post_load, EXCLUDE
from ..resource import Resource
from .customer import Customer
from collections import namedtuple


class Tags(Resource):
    """
    https://dev.chartmogul.com/v1.0/reference#customer-attributes
    """

    _path = "/customers{/uuid}/attributes/tags"

    class _Schema(Schema):
        tags = fields.List(fields.String())

        @post_load
        def make(self, data, **kwargs):
            return Tags(**data)

    _customers = namedtuple("Customers", ["entries"])
    _schema = _Schema(unknown=EXCLUDE)

    @classmethod
    def _load(cls, response):
        """
        Workaround: when using email, the return value is actually
        a list of customers, but without paging.
        """
        response.raise_for_status()
        if response.status_code == 204:
            return None
        jsonObj = response.json()
        if "entries" in jsonObj:
            customers = Customer._schema.load(jsonObj["entries"], many=True)
            return cls._customers(customers)
        else:
            return cls._schema.load(jsonObj)



from datetime import date, datetime
from json import dumps

from promise import Promise
from uritemplate import URITemplate

from .api.config import Config
from .errors import (
    ArgumentMissingError,
    ConfigurationError,
    annotateHTTPError,
    DeprecatedArgumentError,
)
from .retry_request import requests_retry_session

from .version import __version__

"""
HTTP verb mapping. Based on nodejs library.
"""
MAPPINGS = {
    "add": "post",
    "all": "get",
    "cancel": "patch",
    "create": "post",
    "destroy": "delete",
    "merge": "patch",
    "modify": "patch",
    "patch": "patch",
    "remove": "delete",
    "retrieve": "get",
    "update": "put",
    "destroy_with_params": "delete",
    "modify_with_params": "patch",
}

LIST_PARAMS = ["has_more", "summary", "customer_uuid", "data_source_uuid", "cursor"]
ESCAPED_QUERY_KEYS = {"start_date": "start-date", "end_date": "end-date"}


class DataObject(object):
    def __init__(self, **kwargs):
        """
        Any arguments are translated into class attributes.
        """
        self._set_attrs(**kwargs)

    def _set_attrs(self, **kwargs):
        """
        Dynamically creates attributes of the class, so it's not necessary
        to repeat their list besides the serialization schema.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        """
        Pretty-prints any object as
        <ClassName{property='string',number=4,date=datetime.datetime(2003, 8, 4, 21, 41, 43)}>
        """
        return (
            "<"
            + self.__class__.__name__
            + "{"
            + ", ".join(
                [
                    attr + "=" + repr(getattr(self, attr))
                    for attr in sorted(dir(self))
                    if not attr.startswith("_") and not callable(getattr(self, attr))
                ]
            )
            + "}>"
        )


def json_serial(obj):
    """
    JSON serializer for objects not serializable by default json code
    Handles date & datetime.
    """

    if isinstance(obj, date) or isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError("Type not serializable")


class Resource(DataObject):
    _path = ""

    @classmethod
    def _load(cls, response):
        """
        Python doesn't allow arbitrary fields on arrays and having special classes
        for array & cursor attributes is not necessary => namedtuple.
        (immutable, therefore all set on creation by generator)
        """
        response.raise_for_status()
        if response.status_code == 204 or response.status_code == 202:
            return None
        try:
            jsonObj = response.json()
        except ValueError:  # Couldn't parse JSON, probably just text message.
            return response.content
        return cls._loadJSON(jsonObj)

    @classmethod
    def _loadJSON(cls, jsonObj):
        # has load_many capability & is many entries result?
        if "_root_key" in dir(cls) is not None and cls._root_key in jsonObj:
            return cls._many(
                cls._schema.load(jsonObj[cls._root_key], many=True),
                **{key: jsonObj[key] for key in LIST_PARAMS if key in jsonObj},
            )
        else:
            return cls._schema.load(jsonObj)

    @classmethod
    def _preProcessParams(cls, params):
        for key, replacement in ESCAPED_QUERY_KEYS.items():
            if key in params:
                params[replacement] = params[key]
                del params[key]
        return params

    @classmethod
    def _request(cls, config, method, http_verb, path, data=None, **kwargs):
        if http_verb == "get":
            params = cls._preProcessParams(kwargs)
            data = None
        else:
            params = None
            if data is not None:
                data = dumps(data, default=json_serial)

        return (
            Promise(
                lambda resolve, _: resolve(
                    getattr(
                        requests_retry_session(config.max_retries, config.backoff_factor),
                        http_verb,
                    )(
                        config.uri + path,
                        data=data,
                        headers={
                            "content-type": "application/json",
                            "User-Agent": "chartmogul-python/" + __version__,
                        },
                        params=params,
                        auth=config.auth,
                        timeout=config.request_timeout,
                    )
                )
            )
            .then(cls._load)
            .catch(annotateHTTPError)
        )

    @classmethod
    def _expandPath(cls, path, kwargs):
        t = URITemplate(path)
        return t.expand(kwargs)

    @classmethod
    def _validate_arguments(cls, method, kwargs):
        # This enforces user to pass argument, otherwise we could call
        # wrong URL.
        if method in ["destroy", "cancel", "retrieve", "modify", "update"] and "uuid" not in kwargs:
            raise ArgumentMissingError("Please pass 'uuid' parameter")
        if method in ["create", "modify"] and "data" not in kwargs:
            raise ArgumentMissingError("Please pass 'data' parameter")
        if method == "all" and "page" in kwargs:
            raise DeprecatedArgumentError("The 'page' parameter is deprecated")
        if (
            method in ["destroy_all"]
            and "data_source_uuid" not in kwargs
            and "customer_uuid" not in kwargs
        ):
            raise ArgumentMissingError(
                "Please pass 'data_source_uuid' and 'customer_uuid' parameters"
            )
        if method in ["destroy_with_params", "modify_with_params"] and not (
            "id" in kwargs["data"]["subscription_event"]
            or (
                "external_id" in kwargs["data"]["subscription_event"]
                and "data_source_uuid" in kwargs["data"]["subscription_event"]
            )
        ):
            raise ArgumentMissingError(
                "Please pass 'id' parameter or 'data_source_uuid' and 'external_id'"
            )

    @classmethod
    def _method(callerClass, method, http_verb, path=None, useCallerClass=False, useUUIDFor=None):
        @classmethod
        def fc(calleeClass, config, **kwargs):
            if config is None or not isinstance(config, Config):
                raise ConfigurationError(
                    "First argument should be" " instance of chartmogul.Config class!"
                )

            cls = callerClass if useCallerClass else calleeClass

            pathTemp = path  # due to Python closure
            if pathTemp is None:
                pathTemp = cls._path

            cls._validate_arguments(method, kwargs)

            pathTemp = Resource._expandPath(pathTemp, kwargs)

            if useUUIDFor is not None and 'data' in kwargs.keys():
                kwargs["data"][useUUIDFor] = kwargs["uuid"]

            # UUID is always path parameter only.
            if "uuid" in kwargs:
                del kwargs["uuid"]

            return cls._request(config, method, http_verb, pathTemp, **kwargs)

        return fc


def _add_method(cls, method, http_verb, path=None):
    """
    Dynamically define all possible actions.
    """
    fc = Resource._method(method, http_verb, path)
    # Not supported by 2.7, but probably not needed.
    # fc.__doc__ = "Sends %s request to ChartMogul." % http_verb
    # fc.__name__ = method
    setattr(cls, method, fc)


for method, http_verb in MAPPINGS.items():
    _add_method(Resource, method, http_verb)
